package be.iminds.ilabt.jfed.gui_preferences;

import be.iminds.ilabt.jfed.preferences.JFedCorePreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.TerminalFinder;
import be.iminds.ilabt.jfed.util.common.GeniUrn;
import be.iminds.ilabt.jfed.util.common.OSDetector;
import be.iminds.ilabt.jfed.util.common.SshVersionFinder;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import org.apache.commons.configuration.PropertiesConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

/**
 * jFed HighLevel preferences
 */
public class JFedHLPreferences extends JFedCorePreferences {
    private final static Logger LOG = LoggerFactory.getLogger(JFedHLPreferences.class);
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    private static final String DEFAULT_PUTTY_DIR = "PuTTY";
    private static final String[] DEFAULT_PUTTY_LOCATIONS =
            new String[]{System.getenv("ProgramFiles") + File.separator + DEFAULT_PUTTY_DIR,
                         System.getenv("ProgramFiles(x86)") + File.separator + DEFAULT_PUTTY_DIR, "C:\\PuTTY"};


    public JFedHLPreferences() {
        this(true);
    }
    protected JFedHLPreferences(boolean autoSave) {
        super();

        config.setAutoSave(autoSave);
    }
    protected JFedHLPreferences(PropertiesConfiguration propFile, boolean autoSave) {
        super(null, propFile);

        config.setAutoSave(autoSave);
    }

    @Override
    public JFedHLPreferences makeNoSaveCopy() {
        PropertiesConfiguration config = new PropertiesConfiguration();
        config.copy(this.config);
        return new JFedHLPreferences(config, false);
    }


    public static final String DEFAULT_UNIX_SSH_COMMAND = "ssh -A -X -oServerAliveInterval=120 -i '$i' $u@$h -oPort=$p";
    public static final String DEFAULT_UNIX_RDP_COMMAND = "rdesktop -p $P -u $u $h:$p";
    public static final String DEFAULT_WINDOWS_RDP_COMMAND = "mstsc /v:$h:$p";
    public static final String DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_OLD_OPENSSH =
            "ssh -A -X -oServerAliveInterval=120 -i '$i' $u@$h -oPort=$p -oProxyCommand=\"ssh -e none -i '$I' -oPort=$P $U@$H nc -w 5 %h %p\"";
    public static final String DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_NEW_OPENSSH =
            "ssh -A -X -oServerAliveInterval=120 -i '$i' $u@$h -oPort=$p -oProxyCommand=\"ssh -i '$I' -oPort=$P $U@$H -W %h:%p\"";

    protected void fillinBlanks(PropertiesConfiguration config) {
        super.fillinBlanks(config);

        if (OSDetector.getUnixLikeOsFilter().includedOs(OSDetector.os)) {
            if (!config.containsKey(HLPreferenceKey.PREF_COMMAND_RDP.key))
                config.setProperty(HLPreferenceKey.PREF_COMMAND_RDP.key, DEFAULT_UNIX_RDP_COMMAND);

            if (!config.containsKey(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL.key))
                config.setProperty(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL.key, findUnixTerminalCommand());

            boolean autoSshCommand = getBoolean(HLPreferenceKey.PREF_COMMAND_SSH_AUTO, true);

            if (autoSshCommand || !config.containsKey(HLPreferenceKey.PREF_COMMAND_SSH.key))
                config.setProperty(HLPreferenceKey.PREF_COMMAND_SSH.key, DEFAULT_UNIX_SSH_COMMAND);


            if (autoSshCommand ||
                    (!config.containsKey(HLPreferenceKey.PREF_COMMAND_SSH_PROXY.key)) ||
                    Objects.equals(config.getProperty(HLPreferenceKey.PREF_COMMAND_SSH_PROXY.key), DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_OLD_OPENSSH)) {
                /**
                 * OpenSSH version >= 5.4 supports the -W host:port option, which does the netcat trick, without netcat.
                 *
                 *      -W host:port
                 *            Requests that standard input and output on the client be
                 *            forwarded to host on port over the secure channel.  Implies -N,
                 *            -T, ExitOnForwardFailure and ClearAllForwardings.  Works with
                 *            Protocol version 2 only.
                 * */
                boolean newOpenSsh = false;
                //we are on unix, so we call openssh:
                //   ssh -v
                // should return something starting with: OpenSSH_5.3p1 Debian-3ubuntu7.1, OpenSSL 0.9.8k 25 M...
                //  On newer version, it's:   ssh -V
                // so we have to try both


                Boolean supportsNc = SshVersionFinder.checkIfVersionSupportsNc();
                if (supportsNc == null) {
                    LOG.debug("OpenSSH version could not be determined: unknown if -nc option is supported");
                    newOpenSsh = false;
                } else {
                    newOpenSsh = supportsNc;
                }

                LOG.debug("newOpenSsh=" + newOpenSsh + "");

                if (newOpenSsh)
                    config.setProperty(HLPreferenceKey.PREF_COMMAND_SSH_PROXY.key, DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_NEW_OPENSSH);
                else
                    config.setProperty(HLPreferenceKey.PREF_COMMAND_SSH_PROXY.key, DEFAULT_UNIX_PROXY_SSH_COMMAND_FOR_OLD_OPENSSH);
            }
        }


        if (OSDetector.findOS() == OSDetector.OS.WIN) {
            if (!config.containsKey(HLPreferenceKey.PREF_COMMAND_RDP.key))
                config.setProperty(HLPreferenceKey.PREF_COMMAND_RDP.key, DEFAULT_WINDOWS_RDP_COMMAND);

            //try to find settings for PuTTY
            if (!config.containsKey(HLPreferenceKey.PREF_PUTTY_DIRECTORY.key)) {
                //try to detect the location of PuTTY
                String puttyDir = JFedHLPreferences.findPuTTYDirectory();

                boolean enablePageant = true;

                if (puttyDir != null) {
                    config.setProperty(HLPreferenceKey.PREF_PUTTY_DIRECTORY.key, puttyDir);
                    enablePageant = JFedHLPreferences.pageantExists(new File(puttyDir));
                } else {
                    //disable pageant, as we didn't find a suitable PuTTY-dir
                    enablePageant = false;
                }

                if (!config.containsKey(HLPreferenceKey.PREF_SSHAGENT_USE.key))
                    config.setProperty(HLPreferenceKey.PREF_SSHAGENT_USE.key, enablePageant);

            }


        }

        if (!config.containsKey(HLPreferenceKey.PREF_FIRSTRUN_VERSION.key))
            config.setProperty(HLPreferenceKey.PREF_FIRSTRUN_VERSION.key, 0);

        if (!config.containsKey(HLPreferenceKey.PREF_SSHAGENT_USE.key))
            config.setProperty(HLPreferenceKey.PREF_SSHAGENT_USE.key, true);

        if (!config.containsKey(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES.key))
            config.setProperty(HLPreferenceKey.PREF_TERMINAL_KEEP_TMP_FILES.key, true);

        if (!config.containsKey(HLPreferenceKey.PREF_TERMINAL_DEBUG.key))
            config.setProperty(HLPreferenceKey.PREF_TERMINAL_DEBUG.key, true);

        if (!config.containsKey(HLPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED.key))
            config.setProperty(HLPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED.key, true);

        if (!config.containsKey(HLPreferenceKey.PREF_EXPIRATION_WARNING_TIME.key))
            config.setProperty(HLPreferenceKey.PREF_EXPIRATION_WARNING_TIME.key, 60);

        if (!config.containsKey(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED.key))
            config.setProperty(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED.key, true);

        if (!config.containsKey(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME.key))
            config.setProperty(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME.key, 110);
    }

    public boolean isSshAgentAvailable() {
        if (OSDetector.os == OSDetector.OS.WIN) {
            String puttyDir = getString(HLPreferenceKey.PREF_PUTTY_DIRECTORY);
            if (puttyDir == null) return false;
            return new File(puttyDir, "pageant.exe").exists();
        } else {
            //linux or mac
            if (OSDetector.os == OSDetector.OS.MAC)
                return false;
            return true;
        }
    }

    public static boolean isValidPuttyDir(File puttyDir) {
        if (puttyDir == null) return false;
//        LOG.debug("isValidPuttyDir("+puttyDir.getPath()+") exists="+puttyDir.exists()+" isDir="+puttyDir.isDirectory());
        if (!puttyDir.exists()) return false;
        if (!puttyDir.isDirectory()) return false;
        if (!new File(puttyDir, "putty.exe").exists()) return false;
//        if (!new File(puttyDir, "puttygen.exe").exists()) return false;
        if (!new File(puttyDir, "plink.exe").exists()) return false;
        return true;
    }


    public static String findPuTTYDirectory() {
        for (String location : DEFAULT_PUTTY_LOCATIONS) {
            File file = new File(location);
            if (isValidPuttyDir(file))
                return location;
        }
        return null;

    }

    private static boolean pageantExists(File puttyDir) {
        return new File(puttyDir, "pageant.exe").exists();
    }


    private String autoFindUnixTerminalCommand;

    private String findUnixTerminalCommand() {
        if (autoFindUnixTerminalCommand == null) {
            if (OSDetector.os == OSDetector.OS.UNIX)
                autoFindUnixTerminalCommand = TerminalFinder.findLinuxTerminalCommand();

            else if (OSDetector.os == OSDetector.OS.MAC)
                autoFindUnixTerminalCommand = "tell application \"Terminal\" to do script \"%\"";
        }
        if (autoFindUnixTerminalCommand != null) {
            LOG.info("Falling back to auto-detected terminal command: " + autoFindUnixTerminalCommand);
        }
        return autoFindUnixTerminalCommand;

    }

    public boolean isExpirationWarningEnabled() {
        return getBoolean(HLPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED, true);
    }


    public void setExpirationWarningEnabled(boolean enabled) {
        setBoolean(HLPreferenceKey.PREF_EXPIRATION_WARNING_ENABLED, enabled);
    }


    public int getExpirationWarningTime() {
        return getInt(HLPreferenceKey.PREF_EXPIRATION_WARNING_TIME, 60);
    }


    public void setExpirationWarningTime(int value) {
        setInt(HLPreferenceKey.PREF_EXPIRATION_WARNING_TIME, value);
    }


    public boolean isRecoverExpirationWarningEnabled() {
        return getBoolean(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED, true);
    }


    public boolean isTimelineEnabled() {
        return getBoolean(HLPreferenceKey.PREF_TIMELINE_ENABLED, false);
    }


    public void setRecoverExpirationWarningEnabled(boolean enabled) {
        setBoolean(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_ENABLED, enabled);
    }


    public int getRecoverExpirationWarningTime() {
        return getInt(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME, 10);
    }


    public void setRecoverExpirationWarningTime(int value) {
        setInt(HLPreferenceKey.PREF_RECOVER_EXPIRATION_WARNING_TIME, value);
    }


    public void setTimelineEnabled(boolean value) {
        setBoolean(HLPreferenceKey.PREF_TIMELINE_ENABLED, value);
    }

//    public boolean getAddProjectMembersSshKeysByDefault() {
//        return getBoolean(HLPreferenceKey.PREF_ADD_PROJECT_MEMBERS_SSH_KEYS, false);
//    }
//
//    public void setAddProjectMembersSshKeysByDefault(boolean value) {
//        setBoolean(HLPreferenceKey.PREF_ADD_PROJECT_MEMBERS_SSH_KEYS, value);
//    }

    @Nonnull
    private Map<String, Boolean> getExperimentAutoshareMap(@Nonnull GeniUrn projectUrn) {
        String raw = getString(HLPreferenceKey.PREF_EXPERIMENT_AUTOSHARE_JSON, "{}");
        try {
            return new HashMap<>(MAPPER.readValue(raw, Map.class));
        } catch (IOException e) {
            LOG.warn("Error while reading PREF_EXPERIMENT_AUTOSHARE_JSON for \""+projectUrn+"\" (str=\""+raw+"\"", e);
            return new HashMap<>();
        }
    }

    /**
     * @param projectUrn
     * @return null if no auto share settings known. Otherwise, returns whether the experiment should be shared with project members or not.
     */
    @Nullable
    public Boolean getExperimentAutoshare(@Nonnull GeniUrn projectUrn) {
        Map<String, Boolean> mapping = getExperimentAutoshareMap(projectUrn);
        return mapping.get(projectUrn.toString());
    }

    public void setExperimentAutoshare(@Nonnull GeniUrn projectUrn, @Nullable Boolean autoShare) {
        Map<String, Boolean> mapping = getExperimentAutoshareMap(projectUrn);
        if (autoShare == null)
            mapping.remove(projectUrn.toString());
        else
            mapping.put(projectUrn.toString(), autoShare);
        try {
            String s = MAPPER.writeValueAsString(mapping);
            LOG.debug("writing PREF_EXPERIMENT_AUTOSHARE_JSON -> \""+s+"\"");
            setString(HLPreferenceKey.PREF_EXPERIMENT_AUTOSHARE_JSON, s);
        } catch (JsonProcessingException e) {
            LOG.error("Error while writing PREF_EXPERIMENT_AUTOSHARE_JSON after modifying \""+projectUrn+"\": "+mapping, e);
        }
    }
}
